import argparse

import torch
import numpy as np

import ptychi.api as api
from ptychi.api.task import PtychographyTask
from ptychi.utils import get_suggested_object_size, get_default_complex_dtype

import test_utils as tutils


class Test2DPtychoDM(tutils.TungstenDataTester):
    @tutils.TungstenDataTester.wrap_recon_tester(name="test_2d_ptycho_dm")
    def test_2d_ptycho_dm(self):
        self.setup_ptychi(cpu_only=False)

        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0)

        options = api.DMOptions()
        options.data_options.data = data

        options.object_options.initial_guess = torch.ones(
            [1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)],
            dtype=get_default_complex_dtype(),
        )
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True

        options.probe_options.initial_guess = probe
        options.probe_options.power_constraint.probe_power = np.sum(
            np.max(data, axis=-3), axis=(-2, -1)
        )
        options.probe_options.power_constraint.enabled = True
        options.probe_options.optimizable = True

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = False

        options.reconstructor_options.num_epochs = 8
        options.reconstructor_options.chunk_length = 100
        options.reconstructor_options.allow_nondeterministic_algorithms = False
        task = PtychographyTask(options)
        task.run()

        recon = task.get_data_to_cpu("object", as_numpy=True)[0]
        return recon


class Test2DPtychoDMPositionCorrection(tutils.TungstenDataTester):
    @tutils.TungstenDataTester.wrap_recon_tester(name="test_2d_ptycho_dm_position_correction")
    def test_2d_ptycho_dm_position_correction(self):
        self.setup_ptychi(cpu_only=False)

        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0)

        options = api.DMOptions()
        options.data_options.data = data

        options.object_options.initial_guess = torch.ones(
            [1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)],
            dtype=get_default_complex_dtype(),
        )
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True

        options.probe_options.initial_guess = probe
        options.probe_options.power_constraint.probe_power = np.sum(
            np.max(data, axis=-3), axis=(-2, -1)
        )
        options.probe_options.power_constraint.enabled = True
        options.probe_options.optimizable = True

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = True

        options.reconstructor_options.num_epochs = 8
        options.reconstructor_options.chunk_length = 100
        options.reconstructor_options.allow_nondeterministic_algorithms = False
        task = PtychographyTask(options)
        task.run()

        recon = task.get_data_to_cpu("object", as_numpy=True)[0]
        return recon


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--generate-gold", action="store_true")
    args = parser.parse_args()

    tester = Test2DPtychoDM()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_2d_ptycho_dm()

    tester=Test2DPtychoDMPositionCorrection()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_2d_ptycho_dm_position_correction()
